/* ------ */
/* HUBNET */
/* ------ */

identify (Hubnet);

int     NUsers;                         // The number of (regular) stations

TIME    EchoTimeout,
		// The amount of time after which the station assumes that
		// its packet didn't make it through the hub, if the
		// packet's echo hasn't been heard in the meantime.

	HdrRecTime;
		// The amount of time required to recognize the sender
		// of a packet. Used by the transmitter to simulate the 
		// recognition of the echo of its own packet.

long    MinPL, MaxPL, FrameL;           // Packet parameters

	enum    {Idle, Busy};                    // Hub status
	signals {StartEW, ACK, NACK, Timeout};   // Signals

long    TRate;                          // Transmission rate (itus/bit)

/* -------------------- */
/* The Hub station type */
/* -------------------- */
station Hub {

  int   Status;
  Port  **HPorts;       // Array of hub ports
  void  setup ();
};

void  Hub::setup () {
      HPorts = new Port* [NUsers+1];
      // Port 0 is the broadcast port, ports 1..NUsers are selection ports
      // to stations 1..NUsers
      HPorts [0] = create Port (TRate);
      for (int i = 1; i <= NUsers; i++) HPorts [i] = create Port;
      Status = Idle;
};

/* ------------------------------- */
/* The User (regular) station type */
/* ------------------------------- */
station User {

  Port   *SPort,
	 *BPort;

  Packet Buffer;

  void setup () {
      SPort = create Port (TRate);
      BPort = create Port;
  };
};

/* ---------------------------------------------------------- */
/* The  hub  process:  each  selection  port of the hub has a */
/* separate copy of this process                              */
/* ---------------------------------------------------------- */
process HubProcess (Hub) {

  Port  *SPort, *BPort;

  void setup (int pn) {
      SPort = S->HPorts [pn];
      BPort = S->HPorts [0];
  };

  states {Wait, NewPacket, Done};
};

HubProcess::perform {
    state Wait:                                 // Wait for a packet
      SPort->wait (BOT, NewPacket);
    state NewPacket:                            // A packet arrives
      if (S->Status == Busy) skipto (Wait);     // Hub busy, ignore
      S->Status = Busy;                         // Reserve the hub
      BPort->transmit (ThePacket, Done);
    state Done:                                 // Stop packet transmission
      BPort->stop ();
      S->Status = Idle;                         // Release the hub
      proceed (Wait);                           // Wait for more
};

/* ------------------------------------ */
/* The transmitter of a regular station */
/* ------------------------------------ */
process Xmitter (User) {

  Port    *SPort;
  Packet  *Buffer;

  void setup () {
      SPort  = S->SPort;
      Buffer = &(S->Buffer);
  };
  states {NewPacket, Retransmit, Done, Confirmed, Lost};
};

Xmitter::perform {
    state NewPacket:                            // Attempt to acquire a packet
      if (Client->getPacket (Buffer, MinPL, MaxPL, FrameL))
	proceed (Retransmit);
      Client->wait (ARRIVAL, NewPacket);        // No ready packet
    state Retransmit:                           // Transmit or retransmit
      SPort->transmit (Buffer, Done);
      Signal->send (StartEW);                   // Wait for packet echo
      Signal->wait (NACK, Lost);
    state Done:
      SPort->stop ();                           // Terminate the packet
      Signal->wait (NACK, Retransmit);          // Wait for signal from
      Signal->wait (ACK , Confirmed);           // monitor
    state Confirmed:                            // Release the packet
      Buffer->release ();
      proceed (NewPacket);                              // Take care of next one
    state Lost:
      SPort->abort ();
      proceed (Retransmit);
};

/* --------------------------------- */
/* The receiver of a regular station */
/* --------------------------------- */
process Receiver (User) {

  Port  *BPort;

  void setup () {
	BPort = S->BPort;
  };

  states {Wait, NewPacket};

};

Receiver::perform {
    state Wait:                                 // Wait for the end of packet
      BPort->wait (EMP, NewPacket);
    state NewPacket:                            // Accept the packet
      Client->receive (ThePacket, BPort);
      skipto (Wait);                            // and continue
};

process Monitor;

/* -------------------------------------------------------- */
/* An auxiliary alarm clock used by the monitor (see below) */
/* -------------------------------------------------------- */
process Trumpet (User, Monitor) {

  TIME  Delay;

  states {Start, Play};

  void setup (TIME t) {
    Delay = t;
  };
};

Trumpet::perform {
    state Start:                                // After creation
      Timer->wait (Delay, Play);
    state Play:                                 // Play the trumpet
      Signal->send (Timeout);
      terminate ();
};

/* ---------------------------------------------------------- */
/* Detects  the  echo  of  a packet sent by this station or a */
/* timeout                                                    */
/* ---------------------------------------------------------- */
process Monitor (User) {

  Port          *BPort;
  Trumpet       *AClock;

  void setup () {
	BPort = S->BPort;
  };
  states {WaitSignal, WaitEcho, Waiting, NewPacket, Echo, NoEcho};
  perform;
};

Monitor::perform {
    state WaitSignal:                           // Wait for a waking signal
      Signal->wait (StartEW, WaitEcho);
    state WaitEcho:                             // Setup the alarm clock
      AClock = create Trumpet (EchoTimeout);
      proceed (Waiting);
    state Waiting:                              // Continue waiting for echo
      BPort->wait (BOT, NewPacket);
      Signal->wait (Timeout, NoEcho);
    state NewPacket:                            // Determine the packet sender
      if (ThePacket -> Sender == ident (S)) {
	// The  arrival of a packet sent by this station; wait until the
	// packet header (i.e. the sender) is formally recognized
	Timer->wait (HdrRecTime, Echo);
	Signal->wait (Timeout, NoEcho);
      } else
	skipto (Waiting);
    state Echo:                                 // Echo has been recognized
      if (Signal->clear (Timeout) == ERROR)
	// No pending Signal from the Trumpet
	AClock -> terminate ();
      Signal->send (ACK);
      proceed (WaitSignal);
    state NoEcho:                               // Timeout
      Signal->send (NACK);
      proceed (WaitSignal);
};

/* ------------------ */
/* The "Root" process */
/* ------------------ */
process Root {

  states {Start, Exit};
  void  initTopology ();
  void  initTraffic ();
  void  initProtocol ();
};

void    Root::initTopology () {

      long    LinkLength;
      Link    *slk, *blk;
      Hub     *h;
      User    *s1, *s2;

      readIn (NUsers);          // Number of regular stations
      readIn (LinkLength);      // The length of a link segment in itu
      readIn (TRate);           // The transmission rate in bits/itu

      // Create the broadcast link and the hub
      blk = create Link (NUsers+1);
      h = create Hub;
      // Connect the hub to the broadcast link
      h->HPorts [0]->connect (blk);

      // Create selection links (two ports each) and regular stations
      for (int i = 1; i <= NUsers; i++) {
	slk = create Link (2);
	create User;
	((User*)TheStation)->BPort->connect (blk);
	((User*)TheStation)->SPort->connect (slk);
	// Connect the hub to the selection link
	h->HPorts [i]->connect (slk);
      }

      // Set up distances
      for (i = 1; i <= NUsers; i++) {
	// Note that regular stations are numbered 1...NUsers
	s1 = (User*) idToStation (i);
	// Regular station <-> hub
	s1->BPort->setDTo (h->HPorts [0], LinkLength);
	s1->SPort->setDTo (h->HPorts [i], LinkLength);
	for (int j = i+1; j <= NUsers; j++) {
	  // Regular station <-> regular station
	  s2 = (User*) idToStation (j);
	  s1->BPort->setDTo (s2->BPort, LinkLength+LinkLength);
	}
      }
};

void    Root::initTraffic () {

      Traffic  *tr;
      double   MeanMIT, MeanMLE;

      readIn (MinPL);                   // Minimum packet length
      readIn (MaxPL);                   // Maximum packet length
      readIn (FrameL);                  // Frame information (header+trailer)

      readIn (MeanMIT);                 // Mean message interarrival time
      readIn (MeanMLE);                 // Mean message length

      readIn (EchoTimeout);             // Echo waiting time
      readIn (HdrRecTime);              // Header recognition time

      // Create the traffic pattern
      tr = create Traffic (MIT_exp+MLE_exp, MeanMIT, MeanMLE);
      // Add regular stations as senders and receivers
      for (int i = 1; i <= NUsers; i++) {
	tr->addSender (i);
	tr->addReceiver (i);
      }
};

void    Root::initProtocol () {

      for (int i = 1; i <= NUsers; i++) {
	// User stations' processes
	TheStation = idToStation (i);
	create Xmitter;
	create Receiver;
	create Monitor;
      }

      TheStation = idToStation (0);
      // The Hub
      for (i = 1; i <= NUsers; i++)
	create HubProcess, form ("Hub%03d", i), (i);
};

Root::perform {
    state Start:
      setEtu (1000);                    // Time unit = 1000 ITUs
      setLimit (200, 1000000000);       // Set limits
      initTopology ();
      initTraffic ();
      initProtocol ();
      // That's it, now wait until the simulation is over
      Kernel->wait (DEATH, Exit);
    state Exit:
      // Print out simulation results
      idToTraffic (0) -> printPfm ("Global Performance Measures");
      idToLink (0)-> printPfm ("Broadcast Link Performance Measures");
};
